# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe
# - Till Huckeman

"""
Created on Tue Jan  3 15:20:07 2023

@author: till3
"""
from __future__ import annotations

import numpy as np
from jsonschema import validate
from qcodes.parameters import Parameter

from qumada.instrument.buffers import Buffer, BufferException
from qumada.instrument.custom_drivers.Dummies.dummy_dmm import DummyDmm


# %%
class DummyDMMBuffer(Buffer):
    """Buffer for Dummy DMM"""

    AVAILABLE_TRIGGERS: list[str] = ["software"]

    def __init__(self, device: DummyDmm):
        self._device = device
        self._trigger: str | None = None
        self._subscribed_parameters: set[Parameter] = set()
        self._num_points: int | None = None

    def setup_buffer(self, settings: dict) -> None:
        """Sets instrument related settings for the buffer."""

        validate(settings, self.settings_schema)
        self.settings: dict = settings
        self._device.buffer_SR(settings.setdefault("sampling_rate", 512))
        # self._device.buffer_trig_mode("OFF")
        self._set_num_points()  # Method defined below
        # Datapoints to delete at the beginning of dataset due to delay.
        self.delay_data_points = 0
        self.delay = settings.get("delay", 0)
        if self.delay < 0:
            raise BufferException("The Dummy Dac does not support negative delays.")
        else:
            self.delay_data_points = int(self.delay * self._device.buffer_SR())
            self.num_points = self.delay_data_points + self.num_points
            self._device.buffer_n_points(self.num_points)
        self._device.buffer.ready_buffer()

    @property
    def num_points(self) -> int | None:
        return self._num_points

    @num_points.setter
    def num_points(self, num_points) -> None:
        if num_points > 16383:
            raise BufferException(
                "Dummy Dacs Buffer is to small for this measurement. "
                "Please reduce the number of data points or the delay"
            )
        self._num_points = int(num_points)

    def _set_num_points(self) -> None:
        # TODO: Move to parent Buffer Class?
        """
        Calculates number of datapoints and sets
        the num_points accordingly.
        Raises
        ------
        Exception
           Exception if number of points is overdefined.

        Returns
        -------
        None
        """
        if all(k in self.settings for k in ("sampling_rate", "burst_duration", "num_points")):
            raise BufferException("You cannot define sampling_rate, burst_duration and num_points at the same time")
        elif self.settings.get("num_points", False):
            self.num_points = self.settings["num_points"]
        elif all(k in self.settings for k in ("sampling_rate", "burst_duration")):
            self.num_points = int(np.ceil(self.settings["sampling_rate"] * self.settings["burst_duration"]))

    @property
    def trigger(self) -> str | None:
        return self._trigger

    @trigger.setter
    def trigger(self, trigger: str | None) -> None:
        if trigger == "software":
            self._trigger = trigger

    def force_trigger(self) -> None:
        self._device._force_trigger()

    def read_raw(self) -> dict:
        data = {}

        for parameter in self._subscribed_parameters:
            index = self._device.buffer.subscribed_params.index(parameter)
            data[parameter.name] = self._device.buffer.get()[index]
            data[parameter.name] = data[parameter.name][self.delay_data_points : self.num_points]
        data["timestamps"] = np.linspace(0, self.num_points / self._device.buffer_SR(), self.num_points)
        return data

    def read(self) -> dict:
        # TODO: Add timetrace if possible
        return self.read_raw()

    def subscribe(self, parameters: set | list[Parameter]) -> None:
        for parameter in parameters:
            self._device.buffer.subscribe(parameter)
            self._subscribed_parameters.add(parameter)

    def unsubscribe(self, parameters: set | list[Parameter]) -> None:
        for parameter in parameters.copy():
            if parameter in self._device.buffer.subscribed_params:
                self._device.buffer.subscribed_params.remove(parameter)
                self._subscribed_parameters.remove(parameter)
            else:
                print(f"{parameter} is not subscribed")

    def is_subscribed(self, parameter: Parameter) -> bool:
        return parameter in self._subscribed_parameters

    def start(self) -> None:
        self._device.start()

    def stop(self) -> None: ...

    def is_ready(self) -> bool: ...

    def is_finished(self) -> bool:
        return self._device.is_finished()



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe


import json

# pylint: disable=missing-function-docstring
import os
import tempfile
from contextlib import nullcontext as does_not_raise
from datetime import datetime
from random import random

import pytest
from jsonschema import ValidationError
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QApplication, QMessageBox
from pytest_cases import fixture_ref, parametrize
from pytest_mock import MockerFixture
from qcodes.instrument_drivers.mock_instruments import (
    DummyChannelInstrument,
    DummyInstrument,
)
from qcodes.station import Station

import qumada.instrument.mapping as mapping
from qumada.instrument.custom_drivers.Dummies.dummy_dac import DummyDac
from qumada.instrument.custom_drivers.Dummies.dummy_dmm import DummyDmm
from qumada.instrument.mapping import add_mapping_to_instrument, map_terminals_gui
from qumada.instrument.mapping.base import (
    _load_instrument_mapping,
    load_mapped_terminal_parameters,
    save_mapped_terminal_parameters,
)
from qumada.instrument.mapping.Dummies.DummyDac import DummyDacMapping
from qumada.instrument.mapping.mapping_gui import MainWindow
from qumada.measurement.scripts.generic_measurement import Generic_1D_Sweep


@pytest.fixture(name="dmm", scope="session")
def fixture_dmm():
    dmm = DummyDmm("dmm")
    add_mapping_to_instrument(dmm, mapping=mapping.DUMMY_DMM_MAPPING)
    return dmm


@pytest.fixture(name="dac", scope="session")
def fixture_dac():
    dac = DummyDac("dac")
    add_mapping_to_instrument(dac, mapping=DummyDacMapping())
    return dac


@pytest.fixture(name="dci", scope="session")
def fixture_dci():
    dci = DummyChannelInstrument("dci")
    add_mapping_to_instrument(dci, mapping=mapping.DUMMY_CHANNEL_MAPPING)
    return dci


@pytest.fixture(name="station_with_instruments", scope="session")
def fixture_station_with_instruments(dmm, dac, dci):
    station = Station()
    station.add_component(dmm)
    station.add_component(dac)
    station.add_component(dci)
    return station


@pytest.fixture(name="script")
def fixture_script():
    parameters = {
        "dmm": {"voltage": {"type": "gettable"}, "current": {"type": "gettable"}},
        "dac": {
            "voltage": {
                "type": "dynamic",
                "setpoints": [0, 5],
            }
        },
        "T1": {"test_parameter": {"type": "gettable"}},
        "T2": {"test_parameter": {"type": "gettable"}},
    }
    script = Generic_1D_Sweep()
    script.setup(
        parameters=parameters,
        metadata=None,
        add_script_to_metadata=False,
        add_parameters_to_metadata=False,
    )
    return script


@pytest.fixture(name="unmapped_terminal_parameters")
def fixture_unmapped_terminal_parameters():
    terminal_parameters = {
        "dmm": {"voltage": None, "current": None},
        "dac": {"voltage": None},
        "T1": {"test_parameter": None},
        "T2": {"test_parameter": None},
    }
    return terminal_parameters


# TODO: valid_terminal_parameters_mapping, fixture_script, fixture_station_with_instruments is only valid TOGETHER.
# They must exist within some kind of group
@pytest.fixture(name="mapped_terminal_parameters")
def fixture_mapped_terminal_parameters(station_with_instruments):  # valid for given fixture_station_with_instruments
    terminal_params = {
        "dmm": {"voltage": station_with_instruments.dmm.voltage, "current": station_with_instruments.dmm.current},
        "dac": {"voltage": station_with_instruments.dac.ch01.voltage},
        "T1": {"test_parameter": station_with_instruments.dci.A.temperature},
        "T2": {"test_parameter": station_with_instruments.dci.B.temperature},
    }
    return terminal_params


@pytest.fixture
def valid_mapping_data():
    data = {
        "parameter_names": {
            "IDN": "IDN",
            "v1": "voltage",
            "v2": "voltage",
        }
    }
    return data


@pytest.fixture
def invalid_mapping_data():
    data = {
        "parameters": {
            "IDN": "IDN",
            "v1": "voltage",
            "v2": "voltage",
        }
    }
    return data


@pytest.fixture
def valid_mapped_terminal_parameter_data():
    data = {
        "dmm": {
            "voltage": "dmm_voltage",
            "current": "dmm_current",
        },
        "dac": {"voltage": "dac_ch01_voltage"},
        "T1": {"test_parameter": "dci_ChanA_temperature"},
        "T2": {"test_parameter": "dci_ChanB_temperature"},
    }
    return data


@parametrize(
    "mapping_data, expectation",
    [
        (fixture_ref(valid_mapping_data), does_not_raise()),
        (fixture_ref(invalid_mapping_data), pytest.raises(ValidationError)),
    ],
)
def test_load_instrument_mapping(mocker: MockerFixture, mapping_data, expectation):
    with expectation:
        path = "pathtomapping.json"
        mock = mocker.mock_open(read_data=json.dumps(mapping_data))
        mocker.patch("builtins.open", mock)

        ret = _load_instrument_mapping(path)
        assert ret == mapping_data
        mock.assert_called_once_with(path)


def test_save_mapped_terminal_parameters(
    mocker: MockerFixture, mapped_terminal_parameters, valid_mapped_terminal_parameter_data
):
    path = os.path.join(tempfile.gettempdir(), os.urandom(24).hex())
    try:
        save_mapped_terminal_parameters(mapped_terminal_parameters, path)
        with open(path) as file:
            contents = json.load(file)
    finally:
        os.remove(path)
    assert valid_mapped_terminal_parameter_data == contents


def test_load_mapped_terminal_parameters(
    mocker: MockerFixture,
    station_with_instruments: Station,
    unmapped_terminal_parameters,
    mapped_terminal_parameters,
    valid_mapped_terminal_parameter_data,
):
    path = "mapped_terminal_parameters.json"
    mock = mocker.mock_open(read_data=json.dumps(valid_mapped_terminal_parameter_data))
    mocker.patch("builtins.open", mock)

    load_mapped_terminal_parameters(
        terminal_parameters=unmapped_terminal_parameters,
        station=station_with_instruments,
        path=path,
    )
    assert mapped_terminal_parameters == unmapped_terminal_parameters


@pytest.mark.parametrize(
    "path",
    [getattr(mapping, name) for name in dir(mapping) if name.endswith("_MAPPING")],
)
def test_validate_mapping_file(path):
    _load_instrument_mapping(path)


def test_instrument_mapping(mocker: MockerFixture, valid_mapping_data):
    instr = DummyInstrument("instrument", ["v1", "v2"])
    mocker_load = mocker.patch(
        "qumada.instrument.mapping.base._load_instrument_mapping",
        return_value=valid_mapping_data,
    )
    path = "pathtomapping.json"

    add_mapping_to_instrument(instr, mapping=path)

    for parameter in [instr.v1, instr.v2]:
        assert parameter._mapping == "voltage"
    mocker_load.assert_called_once_with(path)

    instr.close()


def test_map_terminals_gui(mocker: MockerFixture, station_with_instruments, script):
    mocker.patch("qumada.instrument.mapping.mapping_gui.QApplication", autospec=True)
    mock_main_window = mocker.patch("qumada.instrument.mapping.mapping_gui.MainWindow", autospec=True)

    kwargs = {
        "components": station_with_instruments.components,
        "terminal_parameters": script.gate_parameters,
        "monitoring": False,
    }
    map_terminals_gui(**kwargs)
    main_window_instance = mock_main_window.return_value

    mock_main_window.assert_called_once_with(**kwargs)
    main_window_instance.show.assert_called_once()


@pytest.mark.parametrize("monitoring", [True, False])
def test_mapping_gui_monitoring(monitoring: bool, qtbot, station_with_instruments, script, mocker):
    w = MainWindow(
        station_with_instruments.components,
        script.gate_parameters,
        monitoring=monitoring,
    )
    w.show()

    assert w.monitoring_enable == monitoring
    assert w.terminal_tree.isColumnHidden(2) != monitoring
    # check menu button label
    if not monitoring:
        assert w.toggle_monitoring_action.text() == "Enable"
    else:
        assert w.toggle_monitoring_action.text() == "Disable"

    # trigger toggle_monitoring_action (this is in the menubar)
    w.toggle_monitoring_action.trigger()

    assert w.monitoring_enable != monitoring
    assert w.terminal_tree.isColumnHidden(2) == monitoring
    if monitoring:
        assert w.toggle_monitoring_action.text() == "Enable"
    else:
        assert w.toggle_monitoring_action.text() == "Disable"

    # test monitoring in action (switch back to enabling monitoring first)
    # This block is a bit convoluted because I am actually measuring the value of the monitored param over time
    # in order to test the monitoring
    #   > this is much nicer if I mock the get command (this somehow didnt work for me...)
    if monitoring:
        w.toggle_monitoring_action.trigger()

        # minimal example (dmm.current get command is random number generator)
        w.terminal_parameters["dmm"]["current"] = station_with_instruments.dmm.current

        # update tree to update visuals with manually set mapping
        w.terminal_tree.update_tree()

        QApplication.processEvents()

        # Test setting the monitoring delay
        # mock QInputDialog.getDouble() dialog (for setting a new rate), set a shorter (but random rate)
        delay_seconds = 0.2 + 0.1 * random()
        mocker.patch("qumada.instrument.mapping.mapping_gui.QInputDialog.getDouble", return_value=(delay_seconds, True))
        w.monitoring_refresh_delay.trigger()

        # sample the last get value, wait, sample, wait, sample (catching the change due to get command in monitoring)
        # careful with setting delay_seconds and the actual waiting times.
        # There might be additional delay in process_events function call and cache.get() making this not
        # completely exact. So dont make the window (time margin) too tight.
        before = station_with_instruments.dmm.current.cache.get()
        # wait 0.5*delay_seconds
        # n = 5  # num_steps
        time_now = datetime.now()
        while (datetime.now() - time_now).microseconds < 1e6 * delay_seconds * 0.5:
            pass
        QApplication.processEvents()
        tmp = station_with_instruments.dmm.current.cache.get()

        # wait another 1 delay_seconds
        time_now = datetime.now()
        while (datetime.now() - time_now).microseconds < 1e6 * delay_seconds:
            pass
        QApplication.processEvents()

        after = station_with_instruments.dmm.current.cache.get()

        # this should not have changed yet
        assert before == tmp

        # this makes sure (with high probability) that monitoring called the get function within the second time window
        assert after != before
        # TODO: maybe test this differently. this can fail with some probability...
        # (new random value is equal to old value)


# This somehow doesnt work with the CI/CD pipeline (inside docker container)
# def test_mapping_gui_map_with_enter(mocker, qtbot, station_with_instruments, script):
#     # mock dialogs (specify behaviour in return_value and skip)
#     mocker.patch("qumada.instrument.mapping.mapping_gui.MessageBox_notallmapped.exec", return_value=QMessageBox.No)
#     mocker.patch("qumada.instrument.mapping.mapping_gui.MessageBox_duplicates.exec", return_value=QMessageBox.No)
#     mocker.patch("qumada.instrument.mapping.mapping_gui.MessageBox_overwrite.exec", return_value=QMessageBox.No)

#     w = MainWindow(
#         station_with_instruments.components,
#         script.gate_parameters,
#     )
#     w.show()
#     qtbot.addWidget(w)
#     for _ in range(8):  # 9 is exactly enough to map all terminals
#         qtbot.keyPress(w, Qt.Key_Return)
#         QApplication.processEvents()

#     # wanted mapping
#     # (TODO: better as fixture? Problem is that this has to be manually set and fit to the specific station fixture
#     # (ORDER) and script fixture)
#     terminal_params = {
#         "dmm": {"voltage": station_with_instruments.dmm.voltage, "current": station_with_instruments.dmm.current},
#         "dac": {"voltage": station_with_instruments.dac.voltage},
#         "T1": {"test_parameter": station_with_instruments.dci.A.temperature},
#         "T2": {"test_parameter": station_with_instruments.dci.B.temperature},
#     }

#     # TODO: assert exact mapping that is expected (create fixture for that and assert equality)
#     assert w.terminal_parameters == terminal_params


def test_mapping_gui_map_automatically(mocker, qtbot, station_with_instruments, script):
    # mock dialogs (specify behaviour in return_value and skip)
    mocker.patch("qumada.instrument.mapping.mapping_gui.MessageBox_notallmapped.exec", return_value=QMessageBox.No)
    mocker.patch("qumada.instrument.mapping.mapping_gui.MessageBox_duplicates.exec", return_value=QMessageBox.No)
    mocker.patch("qumada.instrument.mapping.mapping_gui.MessageBox_overwrite.exec", return_value=QMessageBox.No)

    w = MainWindow(
        station_with_instruments.components,
        script.gate_parameters,
    )
    w.show()
    qtbot.addWidget(w)

    # wanted mapping
    terminal_params = {
        "dmm": {"voltage": station_with_instruments.dmm.voltage, "current": station_with_instruments.dmm.current},
        "dac": {"voltage": station_with_instruments.dac.ch01.voltage},
        "T1": {"test_parameter": station_with_instruments.dci.A.temperature},
        "T2": {"test_parameter": station_with_instruments.dci.B.temperature},
    }

    # check if auto mapping yields wanted result
    # qtbot.mouseClick(w.button_map_auto, Qt.MouseButton.LeftButton)
    qtbot.keyPress(w, Qt.Key_A)
    assert w.terminal_parameters == terminal_params

    # check if exiting works
    # qtbot.mouseClick(w.button_exit, Qt.MouseButton.LeftButton)
    assert w.isVisible()
    qtbot.keyPress(w, Qt.Key_E)
    assert not w.isVisible()



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - 4K User
# - Daniel Grothe
# - Sionludi Lab
# - Till Huckeman


from __future__ import annotations

import json
import logging
from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping, MutableMapping
from typing import Any, Union

import jsonschema
from qcodes.instrument import Instrument
from qcodes.metadatable import Metadatable
from qcodes.parameters import Parameter
from qcodes.station import Station

from qumada.metadata import Metadata

logger = logging.getLogger(__name__)


TerminalParameters = Mapping[Any, Union[Mapping[Any, Parameter], Parameter]]


class MappingError(Exception):
    """Exception is raised, if an error occured during Mapping."""


class InstrumentMapping(ABC):
    def __init__(self, mapping_path: str | None, is_triggerable: bool = False):
        if mapping_path:
            self._mapping = _load_instrument_mapping(mapping_path)
        self._is_triggerable = is_triggerable

    @property
    def mapping(self) -> dict:
        return self._mapping

    @abstractmethod
    def ramp(
        self,
        parameters: list[Parameter],
        *,
        start_values: list[float] | None = None,
        end_values: list[float],
        ramp_time: float,
    ) -> None:
        """Wrapper to ramp the provided parameters"""

    def pulse(
        self,
        parameters: list[Parameter],
        *,
        setpoints: list[float],
        delay: float,
        **kwargs,
    ) -> None:
        """Wrapper to apply pulse of arbitrary values"""

    @abstractmethod
    def setup_trigger_in(self, trigger_settings: dict) -> None:
        """Setup the trigger based on the buffer_settings"""


def filter_flatten_parameters(node) -> dict[Any, Parameter]:
    """
    Recursively filters objects of Parameter types from data structure, that consists of dicts, lists and Metadatable.

    Args:
        node (Union[Dict, List, Metadatable]): Current/starting node in the data structure

    Returns:
        Dict[Any, Parameter]: Flat dict of parameters
    """

    def recurse(node) -> None:
        """Recursive part of the function. Fills instrument_parameters dict."""
        # TODO: Handle InstrumentChannel
        # TODO: Change this try-except-phrase to match-case, when switched to Python3.10
        try:
            values = list(node.values()) if isinstance(node, dict) else list(node)
        except KeyError:
            values = [node]
        except IndexError:
            values = []
        # TODO: Lines 37 and 38 are only a hotfix for problems with the MFLI,
        # The index error is raised somewhere within QCoDeS because the MFLI
        # driver just adds keys that are missing instead of raising the KeyError
        # properly. We should look into this later...
        for value in values:
            if isinstance(value, Parameter):
                instrument_parameters[value.full_name] = value
            else:
                if isinstance(value, Iterable) and not isinstance(value, str):
                    recurse(value)
                elif isinstance(value, Metadatable):
                    # Object of some Metadatable type, try to get __dict__ and _filter_flatten_parameters
                    try:
                        value_hash = hash(value)
                        if value_hash not in seen:
                            seen.add(value_hash)
                            recurse(vars(value))
                    except TypeError:
                        # End of tree
                        pass

    instrument_parameters: dict[Any, Parameter] = {}
    seen: set[int] = set()
    recurse(node)
    return instrument_parameters


def _load_instrument_mapping(path: str) -> Any:
    """
    Loads instrument mapping from mapping JSON file.

    Args:
        path (str): Path to the file.

    Returns:
        Any: Parsed JSON-object
    """
    mapping_structure = {
        "type": "object",
        "properties": {
            "parameter_names": {
                "type": "object",
                "additionalProperties": {"type": "string"},
            }
        },
        "required": ["parameter_names"],
    }

    with open(path) as file:
        mapping_data = json.load(file)
    jsonschema.validate(mapping_data, schema=mapping_structure)
    return mapping_data


def add_mapping_to_instrument(
    instrument: Instrument,
    *,
    mapping: InstrumentMapping | str | None = None,
) -> None:
    """
    Loads instrument mapping from mapping JSON file and adds it as instrument attribute.
    Alternatively, provide a mapping object, that contains the mapping contents, as well as
    wrapper functions e.g. for ramps

    instr._mapping

    Args:
        instrument (Instrument): Instrument, the mapping is added to.
        mapping (InstrumentMapping | str): Either mapping object or path to json file
                                           with mapping.
    """
    if isinstance(mapping, InstrumentMapping):
        helper_mapping = mapping.mapping
        instrument._qumada_ramp = mapping.ramp
        try:
            instrument._qumada_pulse = mapping.pulse
        except TypeError:
            pass
        instrument._is_triggerable = mapping._is_triggerable
        mapping._instrument = instrument
        instrument._qumada_mapping = mapping
        try:
            instrument._qumada_trigger = mapping.trigger
        except Exception:
            pass
    elif isinstance(mapping, str):
        helper_mapping = _load_instrument_mapping(mapping)
        instrument._is_triggerable = False
    else:
        raise ValueError(
            "Mapping parameter has to be either of type \
                         InstrumentMapping or str"
        )

    mapping = {}
    mapping["parameter_names"] = {
        f"{instrument.name}_{key}": parameter for (key, parameter) in helper_mapping["parameter_names"].items()
    }
    parameters: dict[Any, Parameter] = filter_flatten_parameters(instrument)
    mapped_parameters = ((key, parameter) for key, parameter in parameters.items() if key in mapping["parameter_names"])
    for key, parameter in mapped_parameters:
        parameter.__setattr__("_mapping", mapping["parameter_names"][key])


def _generate_mapping_stub(instrument: Instrument, path: str) -> None:
    """
    Generates JSON stub of instrument parametes and saves it under the provided path.
    Overwrites existing files by default.

    The saved JSON-structure is as follows:

    {
        "parameter_names": {
            "instrument_par1": par1,
            "instrument_par2": par2,
            "instrument_par3": par3,
            ...
        }
    }

    After generating the stub, it can be edited, to map different parameter names to the instrument parameters.

    Args:
        instrument (Instrument): Instrument, that shall be parsed
        path (str): Save path for the JSON file
    """
    # Create mapping stub from flat dict of parameters
    mapping = {}
    parameters: dict[Any, Parameter] = filter_flatten_parameters(instrument)
    mapping["parameter_names"] = {
        key.removeprefix(f"{instrument.name}_"): value.name for key, value in parameters.items()
    }

    # Dump JSON file
    with open(path, "w") as file:
        json.dump(mapping, file, indent=4, sort_keys=True)


def map_gates_to_instruments(
    components: Mapping[Any, Metadatable],
    gate_parameters: TerminalParameters,
    existing_gate_parameters: TerminalParameters | None = None,
    *,
    metadata: Metadata | None = None,
    map_manually: bool = False,
) -> None:
    """
    Maps the gates, that were defined in the MeasurementScript to the instruments, that are initialized in QCoDeS.

    Args:
        components (Mapping[Any, Metadatable]): Instruments/Components in QCoDeS
        gate_parameters (TerminalParameters): Gates, as defined in the measurement script
        existing_gate_parameters (TerminalParameters | None): Already existing mapping
                that is used to automatically create the mapping for already known gates without user input.
        metadata (Metadata | None): If provided, add mapping to the metadata object.
        map_manually (bool): If set to True, don't try to automatically map parameters to gates. Defaults to False.
    """
    if existing_gate_parameters is None:
        existing_gate_parameters = {}

    # get all parameters in one flat list for the mapping process
    instrument_parameters = filter_flatten_parameters(components)
    # TODO: We have to distinguish multi channel/module instruments. Possible approach:
    #       [parameter]._instrument should be InstrumentChannel or InstrumentModule type
    for key, gate in gate_parameters.items():
        if isinstance(gate, Parameter):
            # TODO: map single parameter
            # _map_gate_parameters_to_instrument_parameters({key: gate}, )
            ...
        else:
            # map gate to instrument
            # TODO: List is shown even if no user input is required - Fix this
            print(f"Mapping gate {key} to one of the following instruments:")
            for idx, instrument_key in enumerate(components.keys()):
                print(f"{idx}: {instrument_key}")
            chosen = None
            flag = False
            while True:
                try:
                    # Automatically maps all parameters to their corresponding gates
                    # based on the existing mapping
                    for (
                        existing_gate,
                        existing_parameters,
                    ) in existing_gate_parameters.items():
                        if existing_gate == key:
                            if isinstance(existing_parameters, Parameter):
                                # TODO: single parameter
                                ...
                            else:
                                for channel in existing_parameters.values():
                                    if channel:
                                        chosen_instrument = channel.root_instrument
                                        flag = True
                                        print(chosen_instrument)
                                        break

                    # TODO: Does not work with instruments that have only one parameter
                    # (Lists letters of parametername instead of parameter)
                    if not flag:
                        chosen = int(input(f'Which instrument shall be mapped to gate "{key}" ({gate}): '))
                        chosen_instrument = list(components.values())[int(chosen)]
                    chosen_instrument_parameters = {
                        k: v for k, v in instrument_parameters.items() if v.root_instrument is chosen_instrument
                    }
                    try:
                        if map_manually:
                            raise MappingError("map_manually set, mapping manually.")
                        # Only use chosen instrument's parameters for mapping
                        _map_gate_to_instrument(gate, chosen_instrument_parameters)
                        # Remove mapped parameters from parameter list
                        # TODO: remove all parameters from Channel, if parent is a channel
                        keys_to_remove = (
                            key
                            for key in chosen_instrument_parameters.keys()
                            if chosen_instrument_parameters[key] in gate.values()
                        )
                        for key in keys_to_remove:
                            instrument_parameters.pop(key, None)

                    except MappingError as ex:
                        # Could not map instrument, do it manually
                        # TODO: Map to multiple instruments
                        print(ex)
                        _map_gate_parameters_to_instrument_parameters(gate, chosen_instrument_parameters)
                        # Remove mapped parameters from parameter list
                        keys_to_remove = (
                            key
                            for key in chosen_instrument_parameters.keys()
                            if chosen_instrument_parameters[key] in gate.values()
                        )
                        for key in keys_to_remove:
                            instrument_parameters.pop(key, None)
                    break
                except (IndexError, ValueError):
                    continue
    j = json.dumps(gate_parameters, default=lambda o: str(o))
    # Add mapping to metadata, if provided
    if metadata is not None:
        metadata.add_terminal_mapping(json.dumps(j), name="automatic-mapping")


def _map_gate_to_instrument(gate: Mapping[Any, Parameter], instrument_parameters: Mapping[Any, Parameter]) -> None:
    """
    Maps the gate parameters of one specific gate to the parameters of one specific instrument.

    Args:
        gate (Mapping[Any, Parameter]): Gate parameters
        instrument_parameters (Mapping[Any, Parameter]): Instrument parameters available for mapping
    """
    mapped_parameters = {
        key: parameter for key, parameter in instrument_parameters.items() if hasattr(parameter, "_mapping")
    }
    for key, parameter in gate.items():
        # Map only parameters that are not set already
        if parameter is None:
            candidates = [parameter for parameter in mapped_parameters.values() if parameter._mapping == key]
            try:
                gate[key] = candidates.pop(0)
            except IndexError:
                instrument_name = next(iter(instrument_parameters.values())).instrument.name
                raise MappingError(f'No mapping candidate for "{key}" in instrument "{instrument_name}" found.')


def _map_gate_parameters_to_instrument_parameters(
    gate_parameters: Mapping[Any, Parameter],
    instrument_parameters: Mapping[Any, Parameter],
    append_unmapped_parameters=True,
) -> None:
    """
    Maps the gate parameters of one specific gate to the instrument parameters of one specific instrument.

    Args:
        gate_parameters (Mapping[Any, Parameter]): Gate parameters
        instrument_parameters (Mapping[Any, Parameter]): Instrument parameters available for mapping
    """
    mapped_parameters = {
        key: parameter for key, parameter in instrument_parameters.items() if hasattr(parameter, "_mapping")
    }
    unmapped_parameters = {
        key: parameter for key, parameter in instrument_parameters.items() if not hasattr(parameter, "_mapping")
    }

    # This is ugly
    for key, parameter in gate_parameters.items():
        if parameter is None:
            # Filter instrument parameters, if _mapping attribute is equal to key_gp
            # if there is no mapping provided, append those parameters to the list
            # if there are no filtered candidates available, show all parameters
            candidates = {k: p for k, p in mapped_parameters.items() if p._mapping == key}
            if append_unmapped_parameters:
                candidates = candidates | unmapped_parameters
            if not len(candidates):
                candidates = mapped_parameters | unmapped_parameters
            candidates_keys = list(candidates.keys())
            candidates_values = list(candidates.values())
            print("Possible instrument parameters:")
            for idx, candidate_key in enumerate(candidates_keys):
                print(f"{idx}: {candidate_key}")
            chosen = None
            while True:
                try:
                    chosen = int(input(f'Please choose an instrument parameter for gate parameter "{key}": '))
                    gate_parameters[key] = candidates_values[int(chosen)]
                    break
                except (IndexError, ValueError):
                    continue


def save_mapped_terminal_parameters(terminal_parameters: TerminalParameters, path: str) -> None:
    """
    Saves already mapped terminals and components to a json file,
    so they can be loaded easily for the exact same setup.

    The saved JSON-structure is as follows:

    {
        "name_of_parameter": {
            "name_of_terminal": "qcodes_parameter_full_name",
            ...
        },
        "name_of_parameter_2": {
            ...
        },
    }
    """
    # Compile concrete mapping data
    tmp_terminal_parameters = {}
    for terminal_parameter, terminals in terminal_parameters.items():
        if isinstance(terminals, Parameter):
            tmp_terminal_parameters[str(terminal_parameter)] = terminals.full_name
        elif isinstance(terminals, MutableMapping):
            terminals_dict = tmp_terminal_parameters[str(terminal_parameter)] = {}
            for terminal, parameter in terminals.items():
                try:
                    terminals_dict[str(terminal)] = parameter.full_name
                except Exception:
                    logger.warning(
                        "Parameter was not saved: Could not get the 'full_name' of parameter %s.", str(parameter)
                    )
    with open(path, mode="w") as file:
        json.dump(tmp_terminal_parameters, file)


def load_mapped_terminal_parameters(terminal_parameters: TerminalParameters, station: Station, path: str) -> None:
    """
    Loads a concrete mapping, that was previously saved to file.
    If errors occur, the mapping will continue, and a warning will
    be logged.

    Warning: Existing mapping for terminal parameters are overwritten if they are given in the file!
    """
    with open(path) as file:
        tmp_terminal_parameters: Mapping[str, Mapping[str, str] | str] = json.load(file)
        assert isinstance(tmp_terminal_parameters, Mapping)

        for parameter_name, terminals in terminal_parameters.items():
            try:
                assert parameter_name in tmp_terminal_parameters
            except AssertionError:
                logger.warning("Parameter could not be loaded: Parameter %s was not found in file.", parameter_name)

            if not isinstance(terminals, MutableMapping):
                # Single parameter, get component by full name
                try:
                    terminal_parameters[parameter_name] = station.get_component(tmp_terminal_parameters[parameter_name])
                except KeyError:
                    logger.warning(
                        "Parameter could not be loaded: QCoDeS station does not contain parameter %s",
                        tmp_terminal_parameters[parameter_name],
                    )
            else:
                # Parameter terminals
                for terminal_name in terminals.keys():
                    try:
                        assert terminal_name in tmp_terminal_parameters[parameter_name]
                    except AssertionError:
                        logger.warning(
                            "Terminal could not be loaded: Terminal %s_%s was not found in file.",
                            parameter_name,
                            terminal_name,
                        )

                    try:
                        terminals[terminal_name] = station.get_component(
                            tmp_terminal_parameters[parameter_name][terminal_name]
                        )
                    except KeyError:
                        logger.warning(
                            "Parameter could not be loaded: QCoDeS station does not contain parameter %s",
                            tmp_terminal_parameters[parameter_name][terminal_name],
                        )



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Sionludi Lab
# - Till Huckeman
# - Tobias Hangleiter


from __future__ import annotations

import logging
import time
from math import isclose

from qumada.utils.generate_sweeps import generate_sweep

LOG = logging.getLogger(__name__)


class Unsweepable_parameter(Exception):
    pass


def ramp_parameter(
    parameter,
    target,
    ramp_rate: float | None = None,
    ramp_time: float | None = None,
    setpoint_intervall: float = 0.1,
    valid_units: str = "all",
    tolerance: float = 1e-5,
    **kwargs,
):
    """
    Used for ramping float-valued parameters. Allows to specify ramp_rate and/or
    ramp_time. The ramp_time provides an upper limit to the time the sweep may
    take if specified.

    Parameters
    ----------
    parameter : QCoDeS parameter
        Parameter you want to sweep
    target : float
        Target value.
    ramp_rate : float | None, optional
        Specify ramp rate for the sweep. Is only relevant when ramp_time is None
        or ramping is finished before ramp_time has passed. If ramp is to slow,
        ramp_time will be used to define the ramp_speed.The default is None.
    ramp_time : float | None, optional
        Maximum time the ramping may take. If ramp_rate is None, the sweep will
        be performed in ramp_time. Else, it provides an upper limit to the ramp
        time.
    setpoint_intervall : float, optional
        Stepsize of the sweep. The smaller, the smoother the sweep will be
        Very small steps can increase the sweeptime significantly.
        The default is 0.1.
    valid_units : str, optional
        Not used yet. The default is "all".
    tolerance: float, optional
        If abs(current_value- target_value) < tolerance*max(current_value, target_value)
        no ramp is done. Default 1e-5.
    **kwargs : TYPE
        DESCRIPTION.

    Raises
    ------
    Unsweepable_parameter
        Raised when the parameter has non-float values and cannot be swepted.

    Returns
    -------
    BOOL
        True if sweep was completed, False if it failed.

    """
    if parameter._settable is False:
        LOG.warning(f"{parameter} is not _settable and cannot be ramped!")
        return False
    current_value = parameter.get()
    LOG.debug(f"parameter: {parameter}")
    LOG.debug(f"current value: {current_value}")
    LOG.debug(f"ramp rate: {ramp_rate}")
    LOG.debug(f"ramp time: {ramp_time}")

    if isinstance(current_value, float):
        LOG.debug(f"target: {target}")
        if isclose(current_value, target, rel_tol=tolerance):
            LOG.debug("Target value is sufficiently close to current_value, no need to ramp")
            return True

        if not ramp_rate:
            if not ramp_time:
                print("Please specify either ramp_time or ramp_speed")
                return False
            else:
                ramp_rate = abs(current_value - float(target)) / ramp_time

        num_points = int(abs(current_value - float(target)) / (ramp_rate * setpoint_intervall)) + 2
        if ramp_time is not None and ramp_time < abs(current_value - float(target)) / ramp_rate:
            print(
                f"Ramp rate of {parameter} is to low to reach target value in specified"
                "max ramp time. Adapting ramp rate to match ramp time"
            )
            return ramp_parameter(
                parameter=parameter,
                target=target,
                ramp_rate=None,
                ramp_time=ramp_time,
                setpoint_intervall=setpoint_intervall,
                valid_units=valid_units,
                **kwargs,
            )
        sweep = generate_sweep(parameter.get(), target, num_points)
        LOG.debug(f"sweep: {sweep}")
        for value in sweep:
            parameter.set(value)
            time.sleep(setpoint_intervall)
        return True
    else:
        raise Unsweepable_parameter("Parameter has non-float values")
    return False


def ramp_or_set_parameter(
    parameter,
    target,
    ramp_rate: float | None = 0.1,
    ramp_time: float | None = 10,
    setpoint_intervall: float = 0.1,
    **kwargs,
):
    """
    Trys to ramp parameter to specified value, if the parameter values are not
    float, they are just set.
    """
    try:
        ramp_parameter(parameter, target, ramp_rate, ramp_time, setpoint_intervall)
    except Unsweepable_parameter:
        parameter.set(target)



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Till Huckeman
# - Bertha Lab Setup
# - Daniel Grothe
# - Max Lennart Oberländer
# - Sionludi Lab
# - Tobias Hangleiter

import logging
from time import sleep, time

import numpy as np
from qcodes.dataset import dond
from qcodes.dataset.measurements import Measurement
from qcodes.parameters.specialized_parameters import ElapsedTimeParameter

from qumada.instrument.buffers import is_bufferable
from qumada.measurement.doNd_enhanced.doNd_enhanced import (
    _interpret_breaks,
    do1d_parallel_asym,
)
from qumada.measurement.measurement import CustomSweep, MeasurementScript
from qumada.utils.ramp_parameter import ramp_or_set_parameter
from qumada.utils.utils import _validate_mapping, naming_helper

logger = logging.getLogger(__name__)


class Generic_1D_Sweep(MeasurementScript):
    def run(self, **dond_kwargs) -> list:
        """
        Perform 1D sweeps for all dynamic parameters, one after another.

        Dynamic parameters that are not currently active are kept at their
        "value" value.

        Parameters
        ----------
        **dond_kwargs : dict
            Additional keyword arguments passed to the `dond` method.

        Attributes (via settings)
        -------------------------
        wait_time : float, optional
            Wait time (in seconds) between initialization and each measurement.
            Default is 5.
        include_gate_name : bool, optional
            If True, append the name of the ramped gate to the measurement name.
            Default is True.
        ramp_speed : float, optional
            Speed at which parameters are ramped during initialization.
            Default is 0.3.
        ramp_time : float, optional
            Maximum time (in seconds) allowed for ramping each parameter during
            initialization. Default is 10.
        log_idle_params : bool, optional
            If True, record dynamic parameters kept constant during sweeps.
            Default is True.

        Returns
        -------
        list
            A list of QCoDeS datasets for each sweep.
        """
        wait_time = self.settings.get("wait_time", 5)
        include_gate_name = self.settings.get("include_gate_name", True)
        naming_helper(self, default_name="1D Sweep")
        data = list()
        self.generate_lists()
        for sweep, dynamic_parameter in zip(self.dynamic_sweeps, self.dynamic_parameters):
            if include_gate_name:
                self._measurement_name = f"{self.measurement_name} {dynamic_parameter['gate']}"
            else:
                self._measurement_name = self.measurement_name
            if self.settings.get("log_idle_params", True):
                idle_channels = [entry for entry in self.dynamic_channels if entry != sweep.param]
                measured_channels = {*self.gettable_channels, *idle_channels}
            else:
                measured_channels = set(self.gettable_channels)
            inactive_channels = [chan for chan in self.dynamic_channels if chan != sweep.param]
            self.initialize(inactive_dyn_channels=inactive_channels)
            sleep(wait_time)
            data.append(
                dond(
                    sweep,
                    *measured_channels,
                    measurement_name=self._measurement_name,
                    break_condition=_interpret_breaks(self.break_conditions),
                    **dond_kwargs,
                )
            )
        self.clean_up()
        return data


class Generic_nD_Sweep(MeasurementScript):
    def run(self, **dond_kwargs):
        """
        Perform an n-dimensional sweep for n dynamic parameters.

        Parameters
        ----------
        **dond_kwargs : dict
            Additional keyword arguments passed to the `dond` method.

        Attributes (via settings)
        -------------------------
        wait_time : float, optional
            Wait time (in seconds) between initialization and each measurement.
            Default is 5.
        include_gate_name : bool, optional
            If True, append the names of the ramped gates to the measurement name.
            Default is True.
        ramp_speed : float, optional
            Speed at which parameters are ramped during initialization.
            Default is 0.3.
        ramp_time : float, optional
            Maximum time (in seconds) allowed for ramping each parameter during
            initialization. Default is 10.

        Returns
        -------
        QCoDeS dataset
            Dataset containing measurement data.
        """
        self.buffered = False
        self.initialize()
        wait_time = self.settings.get("wait_time", 5)
        include_gate_name = self.settings.get("include_gate_name", True)
        naming_helper(self, default_name="nD Sweep")
        if include_gate_name:
            measurement_name = f"{self.measurement_name} {[gate['gate'] for gate in self.dynamic_parameters]}"
        else:
            try:
                measurement_name = self.measurement_name
            except Exception:
                measurement_name = "measurement"

        for sweep in self.dynamic_sweeps:
            ramp_or_set_parameter(sweep._param, sweep.get_setpoints()[0])
        sleep(wait_time)
        data = dond(
            *tuple(self.dynamic_sweeps),
            *tuple(self.gettable_channels),
            measurement_name=measurement_name,
            break_condition=_interpret_breaks(self.break_conditions),
            use_threads=True,
            **dond_kwargs,
        )
        self.clean_up()
        return data


class Generic_1D_parallel_asymm_Sweep(MeasurementScript):
    def run(self):
        raise Exception(
            "This script was renamed to Generic_1D_parallel_Sweep \
                        and is no longer available. \
                        Please use  Generic_1D_parallel_Sweep instead! \
                        No measurement was started."
        )


class Generic_1D_parallel_Sweep(MeasurementScript):
    """
    Sweeps all dynamic parameters in parallel.

    Supports different sweep rates and setpoints for different parameters.
    All parameters must have the same length.

    Parameters
    ----------
    **kwargs : dict
        Additional keyword arguments:
        - `backsweep_after_break` (bool): Sweeps backwards after a break condition
          is triggered. Default is `False`.
        - `wait_time` (float): Wait time in seconds before starting the sweep. Default is `5`.

    Returns
    -------
    data : Any
        The collected data from the parallel sweep.
    """

    def run(self, **do1d_kwargs):
        naming_helper(self, default_name="Parallel 1D Sweep")
        self.initialize()
        backsweep_after_break = self.settings.get("backsweep_after_break", False)
        wait_time = self.settings.get("wait_time", 5)
        dynamic_params = list()
        for sweep in self.dynamic_sweeps:
            dynamic_params.append(sweep.param)
        sleep(wait_time)
        data = do1d_parallel_asym(
            *tuple(self.gettable_channels),
            param_set=dynamic_params,
            setpoints=[sweep.get_setpoints() for sweep in self.dynamic_sweeps],
            delay=self.dynamic_sweeps[0]._delay,
            measurement_name=self.measurement_name,
            break_condition=_interpret_breaks(self.break_conditions),
            backsweep_after_break=backsweep_after_break,
            **do1d_kwargs,
        )
        self.clean_up()
        return data


class Timetrace(MeasurementScript):
    """
    Timetrace measurement.

    Records data over a specified duration with a given timestep. Note that
    the actual timesteps can vary as the recording time for each datapoint
    is not constant. However, the elapsed time recorded is accurate.

    Parameters
    ----------
    **kwargs : dict
        Additional keyword arguments:
        - `duration` (float): Duration of the measurement in seconds. Default is `300`.
        - `timestep` (float): Time interval between measurements in seconds. Default is `1`.
        - `auto_naming` (bool): If `True`, renames the measurement automatically to "Timetrace". Default is `False`.

    Returns
    -------
    dataset : qcodes.dataset.measurements.Measurement
        The collected dataset containing the recorded measurements.
    """

    def run(self):
        self.initialize(dyn_ramp_to_val=True)
        duration = self.settings.get("duration", 300)
        timestep = self.settings.get("timestep", 1)
        timer = ElapsedTimeParameter("time")
        naming_helper(self, default_name="Timetrace")
        meas = Measurement(name=self.measurement_name)
        meas.register_parameter(timer)
        for parameter in [*self.gettable_channels, *self.dynamic_channels]:
            meas.register_parameter(
                parameter,
                setpoints=[
                    timer,
                ],
            )
        with meas.run() as datasaver:
            timer.reset_clock()
            while timer() < duration:
                now = timer()
                results = [(channel, channel.get()) for channel in [*self.gettable_channels, *self.dynamic_channels]]
                datasaver.add_result((timer, now), *results)
                sleep(timestep)
        dataset = datasaver.dataset
        self.clean_up()
        return dataset


class Timetrace_buffered(MeasurementScript):
    """
    Performs a buffered timetrace measurement.

    The measurement duration and timestep are determined via the buffer settings.
    Dynamic parameters are currently not supported.
    The method does not support "manual" triggering mode, as no ramp is started.
    Using "software" triggering is fine if only one buffered instrument is used;
    otherwise, "hardware" triggering is recommended.
    TODO: Add duration arg

    Parameters
    ----------
    dyn_ramp_to_val : bool, optional
        If True, dynamic parameters are ramped to their respective values during initialization.
    trigger_start : str, optional
        Specifies the trigger start method. Default is "software".
    trigger_reset : callable, optional
        Function to reset the trigger after measurement. Default is None.
    trigger_type : str, optional
        Specifies the type of trigger to use. Can be "software" or "hardware".
        Default is "software".
    auto_naming : bool, optional
        Renames measurement automatically to "Timetrace" if True.

    Returns
    -------
    datasets : list
        A list of QCoDeS datasets containing the measurement results.

    Raises
    ------
    Exception
        If unsupported triggering mode is selected or if a channel cannot be buffered and is not static gettable.

    Notes
    -----
    - Ensure that the buffer settings are properly configured before running the measurement.
    - For "hardware" triggering, the `trigger_start` method must be defined.
    - Dynamic parameters are treated as "static gettable"

    """

    def run(self):
        self.initialize(dyn_ramp_to_val=True)
        # duration = self.settings.get("duration", 300)
        # timestep = self.settings.get("timestep", 1)
        timer = ElapsedTimeParameter("time")
        TRIGGER_TYPES = ["software", "hardware"]
        trigger_start = self.settings.get("trigger_start", "software")  # TODO: this should be set elsewhere
        trigger_reset = self.settings.get("trigger_reset", None)
        trigger_type = _validate_mapping(
            self.settings.get("trigger_type"),
            TRIGGER_TYPES,
            default="software",
            default_key_error="software",
        )
        self.buffered = True
        datasets = []

        self.generate_lists()
        naming_helper(self, default_name="Timetrace")
        meas = Measurement(name=self.measurement_name)

        meas.register_parameter(timer)
        for parameter in [*self.gettable_channels, *self.dynamic_channels]:
            meas.register_parameter(
                parameter,
                setpoints=[
                    timer,
                ],
            )
        # Block required to log gettable and static parameters that are not
        # buffarable (e.g. Dac Channels)
        static_gettables = []
        del_channels = []
        del_params = []
        for parameter, channel in zip(self.gettable_parameters, self.gettable_channels):
            if is_bufferable(channel):
                meas.register_parameter(
                    channel,
                    setpoints=[
                        timer,
                    ],
                )
            elif channel in self.static_channels:
                del_channels.append(channel)
                del_params.append(parameter)
                meas.register_parameter(
                    channel,
                    setpoints=[
                        timer,
                    ],
                )
                parameter_value = self.properties[parameter["gate"]][parameter["parameter"]]["value"]
                static_gettables.append((channel, [parameter_value for _ in range(int(self.buffered_num_points))]))
            else:
                raise Exception(f"{channel} cannot be buffered and is not static gettable")
        for channel in del_channels:
            self.gettable_channels.remove(channel)
        for param in del_params:
            self.gettable_parameters.remove(param)
        for parameter, channel in zip(self.dynamic_parameters, self.dynamic_channels):
            parameter_value = self.properties[parameter["gate"]][parameter["parameter"]]["value"]
            static_gettables.append((channel, [parameter_value for _ in range(int(self.buffered_num_points))]))
        with meas.run() as datasaver:
            # start = timer.reset_clock()
            self.ready_buffers()

            if trigger_type == "manual":
                raise Exception("Manual triggering not supported by Timetrace.")
            if trigger_type == "hardware":
                # Set trigger to high here
                try:
                    trigger_start()
                except Exception:
                    print("Please set a trigger or define a trigger_start method")
                pass

            elif trigger_type == "software":
                for buffer in self.buffers:
                    buffer.force_trigger()

            while not all(buffer.is_finished() for buffer in list(self.buffers)):
                sleep(0.1)
            try:
                trigger_reset()
            except Exception:
                print("No method to reset the trigger defined.")

            results = self.readout_buffers(timestamps=True)
            # TODO: Append values from other dynamic parameters
            datasaver.add_result(
                (timer, results.pop(-1)),
                *results,
                *static_gettables,
            )
            datasets.append(datasaver.dataset)
            self.clean_up()
        return datasets


class Timetrace_with_sweeps(MeasurementScript):
    """
    Performs a timetrace measurement with dynamic sweeps.

    Duration and timestep can be set as keyword arguments, both in seconds.
    Note that the timesteps may vary due to variable recording times per data point.
    The elapsed time recorded is accurate.

    Parameters
    ----------
    duration : int, optional
        Total duration of the measurement in seconds. Default is 300.
    timestep : int, optional
        Time between sweeps in seconds. Default is 1.

    Returns
    -------
    dataset : qcodes.dataset.data_set.DataSet
        A QCoDeS dataset containing the measurement results.

    Notes
    -----
    - Dynamic sweeps are executed at each timestep during the measurement.

    """

    def run(self):
        self.initialize()
        duration = self.settings.get("duration", 300)
        timestep = self.settings.get("timestep", 1)
        # backsweeps = self.settings.get("backsweeps", False)
        timer = ElapsedTimeParameter("time")
        meas = Measurement(name=self.metadata.measurement.name or "timetrace")
        meas.register_parameter(timer)
        setpoints = [timer]
        for parameter in self.dynamic_channels:
            meas.register_parameter(parameter)
            setpoints.append(parameter)
        for parameter in self.gettable_channels:
            meas.register_parameter(parameter, setpoints=setpoints)
        with meas.run() as datasaver:
            timer.reset_clock()
            while timer() < duration:
                for sweep in self.dynamic_sweeps:
                    ramp_or_set_parameter(sweep._param, sweep.get_setpoints()[0], ramp_time=timestep)
                now = timer()
                for i in range(0, len(self.dynamic_sweeps[0].get_setpoints())):
                    for sweep in self.dynamic_sweeps:
                        sweep._param.set(sweep.get_setpoints()[i])
                    set_values = [(sweep._param, sweep.get_setpoints()[i]) for sweep in self.dynamic_sweeps]
                    results = [(channel, channel.get()) for channel in self.gettable_channels]
                    datasaver.add_result((timer, now), *set_values, *results)
                # sleep(timestep)
        dataset = datasaver.dataset
        self.clean_up()
        return dataset


class Timetrace_with_Sweeps_buffered(MeasurementScript):
    """
    Performs a buffered timetrace measurement with dynamic sweeps.

    Duration and timestep are determined via buffer settings.
    This method does not support dynamic parameters and cannot use
    "manual" triggering mode. Use "software" or "hardware" triggers.

    Parameters
    ----------
    duration : int, optional
        Total duration of the measurement in seconds. Default is 300.
    trigger_start : str, optional
        Trigger start method. Default is "software".
    trigger_reset : callable, optional
        Function to reset the trigger after measurement. Default is None.
    sync_trigger : int, optional
        Number of the used sync trigger (QDacs only). Default is None.
    trigger_type : str, optional
        Type of trigger to use ("software", "hardware", "manual"). Default is "software".

    Returns
    -------
    datasets : list
        A list of QCoDeS datasets containing the measurement results.

    Raises
    ------
    AttributeError
        If the required methods for triggering or ramping are not defined.
    TypeError
        If no reset method for the trigger is defined.

    Notes
    -----
    - Dynamic sweeps are executed at each timestep during the measurement.
    """

    def run(self):
        self.initialize()
        duration = self.settings.get("duration", 300)
        # timestep = self.settings.get("timestep", 1)
        timer = ElapsedTimeParameter("time")
        TRIGGER_TYPES = ["software", "hardware", "manual"]
        trigger_start = self.settings.get("trigger_start", "software")  # TODO: this should be set elsewhere
        trigger_reset = self.settings.get("trigger_reset", None)
        sync_trigger = self.settings.get("sync_trigger", None)
        trigger_type = _validate_mapping(
            self.settings.get("trigger_type"),
            TRIGGER_TYPES,
            default="software",
            default_key_error="software",
        )
        self.buffered = True
        datasets = []
        self.generate_lists()
        naming_helper(self, default_name="Timetrace with sweeps")
        meas = Measurement(name=self.measurement_name)
        meas.register_parameter(timer)

        for dynamic_param in self.dynamic_parameters:
            self.properties[dynamic_param["gate"]][dynamic_param["parameter"]]["_is_triggered"] = True
        for dyn_channel in self.dynamic_channels:
            meas.register_parameter(dyn_channel)

        # Block required to log gettable and static parameters that are not
        # buffarable (e.g. Dac Channels)
        static_gettables = []
        for parameter, channel in zip(self.gettable_parameters, self.gettable_channels):
            if is_bufferable(channel) and channel not in self.static_gettable_channels:
                meas.register_parameter(channel, setpoints=[timer, *self.dynamic_channels])
            elif channel in self.static_gettable_channels:
                meas.register_parameter(channel, setpoints=[timer, *self.dynamic_channels])
                parameter_value = self.properties[parameter["gate"]][parameter["parameter"]]["value"]
                static_gettables.append((channel, [parameter_value for _ in range(int(self.buffered_num_points))]))
        start = time()
        with meas.run() as datasaver:
            try:
                trigger_reset()
            except TypeError:
                logger.info("No method to reset the trigger defined.")
            while time() - start < duration:
                self.initialize()
                self.ready_buffers()
                t = time() - start
                try:
                    self.dynamic_channels[0].root_instrument._qumada_ramp(
                        self.dynamic_channels,
                        end_values=[sweep.get_setpoints()[-1] for sweep in self.dynamic_sweeps],
                        ramp_time=self._burst_duration,
                        sync_trigger=sync_trigger,
                    )
                except AttributeError as ex:
                    logger.error(
                        "Exception: This instrument probably does not have a \
                          qtools_ramp method. Buffered measurements without \
                          ramp method are no longer supported. \
                          Use the unbuffered script!"
                    )
                    raise ex

                if trigger_type == "manual":
                    pass
                if trigger_type == "hardware":
                    try:
                        trigger_start()
                    except AttributeError as ex:
                        print("Please set a trigger or define a trigger_start method")
                        raise ex

                elif trigger_type == "software":
                    for buffer in self.buffers:
                        buffer.force_trigger()
                    logger.warning(
                        "You are using software trigger, which \
                        can lead to significant delays between \
                        measurement instruments! Only recommended\
                        for debugging."
                    )
                while not all(buffer.is_finished() for buffer in list(self.buffers)):
                    sleep(0.1)
                try:
                    trigger_reset()
                except TypeError:
                    logger.info("No method to reset the trigger defined.")
                results = self.readout_buffers(timestamps=True)
                dynamic_param_results = [
                    (dyn_channel, sweep.get_setpoints())
                    for dyn_channel, sweep in zip(self.dynamic_channels, self.dynamic_sweeps)
                ]
                results.pop(-1)  # removes timestamps from results
                datasaver.add_result(
                    (timer, t),
                    *dynamic_param_results,
                    *results,
                    *static_gettables,
                )
                self.clean_up()
        datasets.append(datasaver.dataset)
        return datasets


class Generic_1D_Sweep_buffered(MeasurementScript):
    """
    Executes a buffered 1D sweep measurement. Works only with ramps (no pulses). Supports compensation.

    - "software": Sends a software command to each buffer and dynamic parameter to start data acquisition.
                   Timing might have slight offsets.
    - "hardware": Runs trigger_start to start the measurement., either preconfigured or started manually.
    - "manual"  : User-controlled trigger setup, useful for synchronized trigger outputs (e.g., with QDac).

    Parameters
    ----------
    trigger_start : str or callable, optional
        A callable to start the measurement or the keyword "manual" for user-triggered setup.
        Default is "manual".
    trigger_reset : callable, optional
        A callable to reset the trigger after measurement. Default is None.
    trigger_type : str, optional
        Type of trigger to use ("software", "hardware", "manual"). Default is "software".
    include_gate_name : bool, optional
        If True, appends the name of the ramped gates to the measurement name. Default is True.
    sync_trigger : int, optional
        Number of the used sync trigger (QDacs only). Default is None.

    Returns
    -------
    datasets : list of qcodes.dataset.data_set.DataSet
        A list of datasets containing the measurement results.

    Raises
    ------
    AttributeError
        If a required method (e.g., for ramping) is missing.
    TypeError
        If no reset method for the trigger is defined.
    Exception
        If setpoints of a parameter exceed defined limits.

    Notes
    -----
    - This script supports buffered measurements, with dynamic and static parameters.
    - It ensures that all setpoints and compensating channels remain within their defined limits.
    - Proper configuration of buffers and triggers is essential for accurate results.
    """

    def run(self):
        self.buffered = True
        TRIGGER_TYPES = ["software", "hardware", "manual"]
        trigger_start = self.settings.get("trigger_start", "manual")  # TODO: this should be set elsewhere
        trigger_reset = self.settings.get("trigger_reset", None)
        trigger_type = _validate_mapping(
            self.settings.get("trigger_type"),
            TRIGGER_TYPES,
            default="software",
            default_key_error="software",
        )
        include_gate_name = self.settings.get("include_gate_name", True)
        sync_trigger = self.settings.get("sync_trigger", None)
        datasets = []
        self.generate_lists()
        measurement_name = naming_helper(self, default_name="1D Sweep")
        # meas.register_parameter(timer)

        for i in range(len(self.dynamic_sweeps.copy())):
            # dynamic_sweep and dynamic_parameter are from copy and not
            # affected by changes made to parameter in original list!
            self.measurement_name = measurement_name
            dynamic_parameter = self.dynamic_parameters[i]
            if include_gate_name:
                self.measurement_name += f" {dynamic_parameter['gate']}"
            self.properties[dynamic_parameter["gate"]][dynamic_parameter["parameter"]]["_is_triggered"] = True

            dynamic_param = self.dynamic_sweeps[i].param
            inactive_channels = [chan for chan in self.dynamic_channels if chan != dynamic_param]
            self.initialize(inactive_dyn_channels=inactive_channels)
            meas = Measurement(name=self.measurement_name)
            meas.register_parameter(dynamic_param)
            for c_param in self.active_compensating_channels:
                meas.register_parameter(
                    c_param,
                    setpoints=[
                        dynamic_param,
                    ],
                )
            static_gettables = []
            for parameter, channel in zip(self.gettable_parameters, self.gettable_channels):
                if is_bufferable(channel) and channel not in self.static_gettable_channels:
                    meas.register_parameter(
                        channel,
                        setpoints=[
                            dynamic_param,
                        ],
                    )
                elif channel in self.static_gettable_channels:
                    parameter_value = self.properties[parameter["gate"]][parameter["parameter"]]["value"]
                    parameter_value = channel.get()
                    static_gettables.append((channel, [parameter_value for _ in range(int(self.buffered_num_points))]))
            for parameter, channel in zip(self.dynamic_parameters, self.dynamic_channels):
                if channel != dynamic_param:
                    try:
                        parameter_value = self.properties[parameter["gate"]][parameter["parameter"]]["value"]
                    except KeyError:
                        logger.error(
                            "An idle dynamic parameter has no value assigned\
                              and cannot be logged!"
                        )
                        break
                    static_gettables.append((channel, [parameter_value for _ in range(int(self.buffered_num_points))]))
            for param in static_gettables:
                meas.register_parameter(
                    param[0],
                    setpoints=[
                        dynamic_param,
                    ],
                )
            active_comping_sweeps = []
            for j in range(len(self.active_compensating_channels)):
                index = self.compensating_parameters.index(self.active_compensating_parameters[j])
                active_comping_setpoints = self.compensating_parameters_values[index] + sum(
                    [sweep.get_setpoints() for sweep in self.compensating_sweeps[j]]
                )
                if min(active_comping_setpoints) < min(self.compensating_limits[index]) or max(
                    active_comping_setpoints
                ) > max(self.compensating_limits[index]):
                    raise Exception(f"Setpoints of {self.compensating_parameters[index]} exceed limits!")
                sweep_delay = self.compensating_sweeps[j][-1]._delay
                active_comping_sweeps.append(
                    CustomSweep(
                        param=self.active_compensating_channels[j],
                        setpoints=active_comping_setpoints,
                        delay=sweep_delay,
                    )
                )

            meas.write_period = 0.5

            with meas.run() as datasaver:

                dynamic_sweep = self.dynamic_sweeps[i]
                try:
                    trigger_reset()
                except TypeError:
                    logger.info("No method to reset the trigger defined.")
                results = []
                self.ready_buffers()
                try:
                    dynamic_param.root_instrument._qumada_ramp(
                        [dynamic_param, *self.active_compensating_channels],
                        end_values=[
                            dynamic_sweep.get_setpoints()[-1],
                            *[sweep.get_setpoints()[-1] for sweep in active_comping_sweeps],
                        ],
                        ramp_time=self._burst_duration,
                        sync_trigger=sync_trigger,
                    )
                except AttributeError as ex:
                    logger.error(
                        "Exception: This instrument probably does not have a \
                          a qumada_ramp method. Buffered measurements without \
                          ramp method are no longer supported. \
                          Use the unbuffered script!"
                    )
                    raise ex

                if trigger_type == "manual":
                    pass
                if trigger_type == "hardware":
                    try:
                        trigger_start()
                    except AttributeError as e:
                        print("Please set a trigger or define a trigger_start method")
                        raise e

                elif trigger_type == "software":
                    for buffer in self.buffers:
                        buffer.force_trigger()
                    logger.warning(
                        "You are using software trigger, which \
                        can lead to significant delays between \
                        measurement instruments! Only recommended \
                        for debugging."
                    )
                while not all(buffer.is_finished() for buffer in list(self.buffers)):
                    sleep(0.1)
                try:
                    trigger_reset()
                except TypeError:
                    logger.info("No method to reset the trigger defined.")

                results = self.readout_buffers()
                comp_results = []
                for ch, sw in zip(self.active_compensating_channels, active_comping_sweeps):
                    comp_results.append((ch, sw.get_setpoints()))
                datasaver.add_result(
                    (dynamic_param, dynamic_sweep.get_setpoints()),
                    *comp_results,
                    *results,
                    *static_gettables,
                )
                datasets.append(datasaver.dataset)
                self.properties[dynamic_parameter["gate"]][dynamic_parameter["parameter"]]["_is_triggered"] = False
                self.clean_up()
        return datasets


class Generic_1D_Hysteresis_buffered(MeasurementScript):
    """
    Performs a buffered 1D hysteresis measurement (back and foresweeps).
    Each iteration corresponds to one fore- and one backsweep.

    This measurement supports dynamic and static parameters and handles multiple
    triggering methods:

    - "software": Sends a software command to each buffer and dynamic parameters
                   to start data acquisition and ramping. Timing might be slightly off.
    - "hardware": Runs trigger_start to start the measurement.. Can be preconfigured
                   or manually adjusted (requires `trigger_start` callable).
    - "manual": Trigger setup is user-defined, useful for synchronized trigger outputs.

    Parameters
    ----------
    iterations : int
        Defines how many sweeps are done. Each iteration corresponds to one fore- and one backsweep.
    trigger_start : str or callable, optional
        A callable to start the measurement or "manual" for user-triggered setup.
        Default is "manual".
    trigger_reset : callable, optional
        A callable to reset the trigger after measurement. Default is None.
    trigger_type : str, optional
        Type of trigger to use ("software", "hardware", "manual"). Default is "software".
    sync_trigger : int, optional
        Number of the used sync trigger (QDacs only). Default is None.
    include_gate_name : bool, optional
        If True, appends the name of the ramped gates to the measurement name. Default is True.

    Returns
    -------
    datasets : list of qcodes.dataset.data_set.DataSet
        A list of datasets containing the measurement results.

    Raises
    ------
    AttributeError
        If a required method (e.g., for ramping) is missing.
    TypeError
        If no reset method for the trigger is defined.
    Exception
        If static or dynamic parameters have invalid configurations.

    Notes
    -----
    - This script performs back-and-forth sweeps (hysteresis) for dynamic parameters.
    - Proper configuration of buffers and triggers is required for accurate results.
    """

    def run(self):
        self.buffered = True
        TRIGGER_TYPES = ["software", "hardware", "manual"]
        trigger_start = self.settings.get("trigger_start", "manual")  # TODO: this should be set elsewhere
        trigger_reset = self.settings.get("trigger_reset", None)
        trigger_type = _validate_mapping(
            self.settings.get("trigger_type"),
            TRIGGER_TYPES,
            default="software",
            default_key_error="software",
        )
        include_gate_name = self.settings.get("include_gate_name", True)
        sync_trigger = self.settings.get("sync_trigger", None)
        iterations = self.settings.get("iterations", 1)
        iterations *= 2
        datasets = []
        self.generate_lists()
        measurement_name = naming_helper(self, default_name="1D Sweep")
        # meas.register_parameter(timer)
        for dynamic_sweep, dynamic_parameter in zip(self.dynamic_sweeps.copy(), self.dynamic_parameters.copy()):
            self.measurement_name = measurement_name
            if include_gate_name:
                self.measurement_name += f" {dynamic_parameter['gate']}"
            self.properties[dynamic_parameter["gate"]][dynamic_parameter["parameter"]]["_is_triggered"] = True
            dynamic_param = dynamic_sweep.param
            meas = Measurement(name=self.measurement_name)
            meas.register_parameter(dynamic_param)
            # This next block is required to log static and idle dynamic
            # parameters that cannot be buffered.
            static_gettables = []
            del_channels = []
            del_params = []
            for parameter, channel in zip(self.gettable_parameters, self.gettable_channels):
                if is_bufferable(channel):
                    meas.register_parameter(
                        channel,
                        setpoints=[
                            dynamic_param,
                        ],
                    )
                elif channel in self.static_channels:
                    del_channels.append(channel)
                    del_params.append(parameter)
                    parameter_value = self.properties[parameter["gate"]][parameter["parameter"]]["value"]
                    static_gettables.append((channel, [parameter_value for _ in range(int(self.buffered_num_points))]))
            for parameter, channel in zip(self.dynamic_parameters, self.dynamic_channels):
                if channel != dynamic_param:
                    try:
                        parameter_value = self.properties[parameter["gate"]][parameter["parameter"]]["value"]
                    except KeyError:
                        logger.error(
                            "An idle dynamic parameter has no value assigned\
                              and cannot be logged!"
                        )
                        break
                    static_gettables.append((channel, [parameter_value for _ in range(int(self.buffered_num_points))]))
            for param in static_gettables:
                meas.register_parameter(
                    param[0],
                    setpoints=[
                        dynamic_param,
                    ],
                )
            for channel in del_channels:
                self.gettable_channels.remove(channel)
            for param in del_params:
                self.gettable_parameters.remove(param)

            try:
                trigger_reset()
            except TypeError:
                logger.info("No method to reset the trigger defined.")

            with meas.run() as datasaver:
                inactive_channels = [chan for chan in self.dynamic_channels if chan != dynamic_param]
                self.initialize(inactive_dyn_channels=inactive_channels)
                results = []

                for iiter in range(0, iterations):
                    self.ready_buffers()
                    if iiter % 2 == 0:
                        set_points = dynamic_sweep.get_setpoints()
                        end_value = dynamic_sweep.get_setpoints()[-1]

                    else:
                        end_value = dynamic_sweep.get_setpoints()[0]
                        set_points = list(reversed(dynamic_sweep.get_setpoints()))
                    try:
                        dynamic_param.root_instrument._qumada_ramp(
                            [dynamic_param],
                            start_values=None,
                            end_values=[end_value],
                            ramp_time=self._burst_duration,
                            sync_trigger=sync_trigger,
                        )
                    except AttributeError as ex:
                        logger.error(
                            "Exception: This instrument probably does not have a \
                              a qumada_ramp method. Buffered measurements without \
                              ramp method are no longer supported. \
                              Use the unbuffered script!"
                        )
                        raise ex

                    if trigger_type == "manual":
                        pass
                    if trigger_type == "hardware":
                        try:
                            trigger_start()
                        except AttributeError as ex:
                            print("Please set a trigger or define a trigger_start method")
                            raise ex

                    elif trigger_type == "software":
                        for buffer in self.buffers:
                            buffer.force_trigger()
                        logger.info(
                            "You are using software trigger, which \
                                        can lead to significant delays between \
                                        measurement instruments! Only recommended\
                                        for debugging."
                        )
                    while not all(buffer.is_finished() for buffer in list(self.buffers)):
                        sleep(0.1)
                    try:
                        trigger_reset()
                    except TypeError:
                        logger.info("No method to reset the trigger defined.")

                    results = self.readout_buffers()
                    datasaver.add_result(
                        (dynamic_param, set_points),
                        *results,
                        *static_gettables,
                    )
                datasets.append(datasaver.dataset)
                self.properties[dynamic_parameter["gate"]][dynamic_parameter["parameter"]]["_is_triggered"] = False
                self.clean_up()
        return datasets


class Generic_2D_Sweep_buffered(MeasurementScript):
    """
    Executes a buffered 2D sweep measurement. Supports compensation.
    By default fist dynamic parameter is stepped (unbuffered) and the second one
    ramped.

    This script supports two dynamic parameters and multiple triggering methods:

    - "software": Sends a software command to each buffer and dynamic parameters
                   to start data acquisition and ramping. Timing might be slightly off.
    - "hardware": Runs trigger_start to start the measurement.. Can be preconfigured
                   or manually adjusted (requires `trigger_start` callable).
    - "manual": Trigger setup is user-defined, useful for synchronized trigger outputs.

    Parameters
    ----------
    trigger_start : str or callable, optional
        A callable to start the measurement.
    trigger_reset : callable, optional
        A callable to reset the trigger after measurement. Default is None.
    trigger_type : str, optional
        Type of trigger to use ("software", "hardware", "manual"). Default is "software".
    include_gate_name : bool, optional
        If True, appends the names of the ramped gates to the measurement name. Default is True.
    reset_time : float, optional
        Time to ramp the fast parameter back to the start value. Default is 0.
    reverse_param_order : bool, optional
        If True, switches the order of slow and fast parameters. Default is False.
    buffer_timeout_multiplier : int, optional
        Multiplier for buffer timeout duration relative to burst duration. Default is 20.

    Returns
    -------
    datasets : list of qcodes.dataset.data_set.DataSet
        A list of datasets containing the measurement results.

    Raises
    ------
    AttributeError
        If a required method (e.g., for ramping) is missing.
    TimeoutError
        If buffers fail to finish within the timeout duration.
    Exception
        If static or dynamic parameters have invalid configurations.

    Notes
    -----
    - Proper configuration of buffers and triggers is required for accurate results.
    """

    def run(self):
        self.buffered = True
        TRIGGER_TYPES = ["software", "hardware", "manual"]
        trigger_start = self.settings.get("trigger_start", "manual")  # TODO: this should be set elsewhere
        trigger_reset = self.settings.get("trigger_reset", None)
        trigger_type = _validate_mapping(
            self.settings.get("trigger_type"),
            TRIGGER_TYPES,
            default="software",
            default_key_error="software",
        )
        include_gate_name = self.settings.get("include_gate_name", True)
        sync_trigger = self.settings.get("sync_trigger", None)
        reverse_param_order = self.settings.get("reverse_param_order", False)
        reset_time = self.settings.get("reset_time", 0)
        buffer_timeout_multiplier = self.settings.get("buffer_timeout_multiplier", 20)
        datasets = []

        self.generate_lists()

        if len(self.dynamic_sweeps) != 2:
            raise Exception("The 2D workflow takes exactly two dynamic parameters! ")
        self.measurement_name = naming_helper(self, default_name="2D Sweep")
        if include_gate_name:
            gate_names = [gate["gate"] for gate in self.dynamic_parameters]
            self.measurement_name += f" {gate_names}"

        meas = Measurement(name=self.measurement_name)

        if reverse_param_order:
            slow_param = self.dynamic_parameters[1]
            slow_channel = self.dynamic_channels[1]
            slow_sweep = self.dynamic_sweeps[1]
            fast_param = self.dynamic_parameters[0]
            fast_channel = self.dynamic_channels[0]
            fast_sweep = self.dynamic_sweeps[0]
            self.properties[self.dynamic_parameters[0]["gate"]][self.dynamic_parameters[0]["parameter"]][
                "_is_triggered"
            ] = True
        else:
            slow_param = self.dynamic_parameters[0]
            slow_channel = self.dynamic_channels[0]
            slow_sweep = self.dynamic_sweeps[0]
            fast_param = self.dynamic_parameters[1]
            fast_channel = self.dynamic_channels[1]
            fast_sweep = self.dynamic_sweeps[1]
            self.properties[self.dynamic_parameters[1]["gate"]][self.dynamic_parameters[1]["parameter"]][
                "_is_triggered"
            ] = True

        for dynamic_param in self.dynamic_channels:
            meas.register_parameter(dynamic_param)
        # -------------------
        static_gettables = []
        del_channels = []
        del_params = []
        for parameter, channel in zip(self.gettable_parameters, self.gettable_channels):
            if is_bufferable(channel):
                meas.register_parameter(
                    channel,
                    setpoints=[
                        slow_channel,
                        fast_channel,
                    ],
                )
            elif channel in self.static_channels:
                del_channels.append(channel)
                del_params.append(parameter)
                meas.register_parameter(
                    channel,
                    setpoints=[
                        slow_channel,
                        fast_channel,
                    ],
                )
                parameter_value = self.properties[parameter["gate"]][parameter["parameter"]]["value"]
                static_gettables.append((channel, [parameter_value for _ in range(int(self.buffered_num_points))]))
        for channel in del_channels:
            self.gettable_channels.remove(channel)
        for param in del_params:
            self.gettable_parameters.remove(param)
        # --------------------------
        self.initialize()
        # ####################Sensor compensation#####################
        for c_param in self.active_compensating_channels:
            meas.register_parameter(
                c_param,
                setpoints=[
                    slow_channel,
                    fast_channel,
                ],
            )
        try:
            trigger_reset()
        except TypeError:
            logger.info("No method to reset the trigger defined.")
        with meas.run() as datasaver:
            results = []
            slow_setpoints = slow_sweep.get_setpoints()
            for setpoint in slow_setpoints:
                slow_channel.set(setpoint)
                if reset_time > 0:
                    ramp_or_set_parameter(
                        fast_channel, fast_sweep.get_setpoints()[0], ramp_rate=None, ramp_time=reset_time
                    )
                else:
                    fast_channel.set(fast_sweep.get_setpoints()[0])
                if reset_time < slow_sweep._delay:
                    sleep(slow_sweep._delay - reset_time)

                comping_results = []
                active_comping_sweeps = []
                for j in range(len(self.active_compensating_channels)):
                    index = self.compensating_parameters.index(self.active_compensating_parameters[j])
                    active_comping_setpoints = np.array(
                        [self.compensating_parameters_values[index] for _ in range(len(fast_sweep.get_setpoints()))],
                        dtype=float,
                    )
                    try:
                        slow_index = self.compensated_parameters[j].index(slow_param)
                        active_comping_setpoints -= float(self.compensating_leverarms[j][slow_index]) * (
                            float(setpoint) - float(slow_sweep.get_setpoints()[0])
                        )
                    except ValueError:
                        pass
                    try:
                        fast_index = self.compensated_parameters[j].index(fast_param)
                        active_comping_setpoints += self.compensating_sweeps[j][fast_index].get_setpoints()
                    except ValueError:
                        pass

                    if min(active_comping_setpoints) < min(self.compensating_limits[index]) or max(
                        active_comping_setpoints
                    ) > max(self.compensating_limits[index]):
                        raise Exception(f"Setpoints of {self.compensating_parameters[index]} exceed limits!")
                    sweep_delay = self.compensating_sweeps[j][-1]._delay
                    active_comping_sweeps.append(
                        CustomSweep(
                            param=self.active_compensating_channels[j],
                            setpoints=active_comping_setpoints,
                            delay=sweep_delay,
                        )
                    )
                    comping_results.append((self.active_compensating_channels[j], active_comping_setpoints))

                self.ready_buffers()
                try:
                    fast_channel.root_instrument._qumada_ramp(
                        [fast_channel, *self.active_compensating_channels],
                        start_values=[
                            fast_sweep.get_setpoints()[0],
                            *[sweep.get_setpoints()[0] for sweep in active_comping_sweeps],
                        ],
                        end_values=[
                            fast_sweep.get_setpoints()[-1],
                            *[sweep.get_setpoints()[-1] for sweep in active_comping_sweeps],
                        ],
                        ramp_time=self._burst_duration,
                        sync_trigger=sync_trigger,
                    )
                except AttributeError as ex:
                    logger.error(
                        "Exception: This instrument probably does not have a \
                          a qumada_ramp method. Buffered measurements without \
                          ramp method are no longer supported. \
                          Use the unbuffered script!"
                    )
                    raise ex

                if trigger_type == "manual":
                    pass

                if trigger_type == "hardware":
                    try:
                        trigger_start()
                    except NameError as ex:
                        print("Please set a trigger or define a trigger_start method")
                        raise ex

                elif trigger_type == "software":
                    for buffer in self.buffers:
                        buffer.force_trigger()
                    logger.warning(
                        "You are using software trigger, which \
                        can lead to significant delays between \
                        measurement instruments! Only recommended\
                        for debugging."
                    )
                timer = 0
                while not all(buffer.is_finished() for buffer in list(self.buffers)):
                    timer += 0.1
                    sleep(0.1)
                    if timer >= buffer_timeout_multiplier * self._burst_duration:
                        raise TimeoutError
                try:
                    trigger_reset()
                except TypeError:
                    logger.info(
                        "No method to reset the trigger defined. \
                        As you are doing a 2D Sweep, this can have undesired \
                        consequences!"
                    )

                results = self.readout_buffers()
                datasaver.add_result(
                    (slow_channel, setpoint),
                    (fast_channel, fast_sweep.get_setpoints()),
                    *comping_results,
                    *results,
                    *static_gettables,
                )
        datasets.append(datasaver.dataset)
        self.clean_up()
        return datasets


class Generic_Pulsed_Measurement(MeasurementScript):
    """
    Executes a buffered pulsed measurement with arbitrary setpoints.
    All setpoint arrays have to have the same length!

    Triggering methods:

    - "software": Sends a software command to each buffer and dynamic parameters
                   to start data acquisition and ramping. Timing might be slightly off.
    - "hardware": Runs trigger_start to start the measurement.. Can be preconfigured
                   or manually adjusted (requires `trigger_start` callable).
    - "manual": Trigger setup is user-defined, useful for synchronized trigger outputs.

    Parameters
    ----------
    trigger_start : str or callable, optional
        A callable to start the measurement or "manual" for user-triggered setup.
        Default is "manual".
    trigger_reset : callable, optional
        A callable to reset the trigger after measurement. Default is None.
    trigger_type : str, optional
        Type of trigger to use ("software", "hardware", "manual"). Default is "software".
    include_gate_name : bool, optional
        If True, appends the name of the ramped gates to the measurement name. Default is True.
    buffer_timeout_multiplier : int, optional
        Multiplier for buffer timeout duration relative to burst duration. Default is 20.
    sync_trigger : callable, optional
        Method for synchronized triggering. Default is None.

    Returns
    -------
    datasets : list of qcodes.dataset.data_set.DataSet
        A list of datasets containing the measurement results.

    Raises
    ------
    AttributeError
        If a required method (e.g., for pulsing) is missing.
    TimeoutError
        If buffers fail to finish within the timeout duration.
    Exception
        If static or dynamic parameters have invalid configurations.

    """

    def run(self):
        self.buffered = True
        TRIGGER_TYPES = ["software", "hardware", "manual"]
        trigger_start = self.settings.get("trigger_start", "manual")  # TODO: this should be set elsewhere
        trigger_reset = self.settings.get("trigger_reset", None)
        trigger_type = _validate_mapping(
            self.settings.get("trigger_type"),
            TRIGGER_TYPES,
            default="software",
            default_key_error="software",
        )
        buffer_timeout_multiplier = self.settings.get("buffer_timeout_multiplier", 20)
        include_gate_name = self.settings.get("include_gate_name", True)
        sync_trigger = self.settings.get("sync_trigger", None)
        datasets = []
        timer = ElapsedTimeParameter("time")
        self.generate_lists()
        self.measurement_name = naming_helper(self, default_name="nD Sweep")
        if include_gate_name:
            gate_names = [gate["gate"] for gate in self.dynamic_parameters]
            self.measurement_name += f" {gate_names}"

        meas = Measurement(name=self.measurement_name)
        meas.register_parameter(timer)
        for parameter in self.dynamic_parameters:
            self.properties[parameter["gate"]][parameter["parameter"]]["_is_triggered"] = True
        for dynamic_param in self.dynamic_channels:
            meas.register_parameter(
                dynamic_param,
                setpoints=[
                    timer,
                ],
            )

        # -------------------
        static_gettables = []
        del_channels = []
        del_params = []
        for parameter, channel in zip(self.gettable_parameters, self.gettable_channels):
            if is_bufferable(channel):
                meas.register_parameter(
                    channel,
                    setpoints=[
                        timer,
                    ],
                )
            elif channel in self.static_channels:
                del_channels.append(channel)
                del_params.append(parameter)
                meas.register_parameter(
                    channel,
                    setpoints=[
                        timer,
                    ],
                )
                parameter_value = self.properties[parameter["gate"]][parameter["parameter"]]["value"]
                static_gettables.append((channel, [parameter_value for _ in range(self.buffered_num_points)]))
        for channel in del_channels:
            self.gettable_channels.remove(channel)
        for param in del_params:
            self.gettable_parameters.remove(param)
        # --------------------------

        self.initialize()
        for c_param in self.active_compensating_channels:
            meas.register_parameter(
                c_param,
                setpoints=[
                    timer,
                ],
            )

        instruments = {param.root_instrument for param in self.dynamic_channels}
        time_setpoints = np.linspace(0, self._burst_duration, int(self.buffered_num_points))
        setpoints = [sweep.get_setpoints() for sweep in self.dynamic_sweeps]
        compensating_setpoints = []
        for i in range(len(self.active_compensating_channels)):
            index = self.compensating_channels.index(self.active_compensating_channels[i])
            active_setpoints = sum([sweep.get_setpoints() for sweep in self.compensating_sweeps[i]])
            active_setpoints += float(self.compensating_parameters_values[index])
            compensating_setpoints.append(active_setpoints)
            if min(active_setpoints) < min(self.compensating_limits[index]) or max(active_setpoints) > max(
                self.compensating_limits[index]
            ):
                raise Exception(f"Setpoints of compensating gate {self.compensating_parameters[index]} exceed limits!")
        try:
            trigger_reset()
        except TypeError:
            logger.info("No method to reset the trigger defined.")
        with meas.run() as datasaver:
            results = []
            self.ready_buffers()
            for instr in instruments:
                try:
                    instr._qumada_pulse(
                        parameters=[*self.dynamic_channels, *self.active_compensating_channels],
                        setpoints=[*setpoints, *compensating_setpoints],
                        delay=self._burst_duration / self.buffered_num_points,
                        sync_trigger=sync_trigger,
                    )
                except AttributeError as ex:
                    logger.error(
                        f"Exception: {instr} probably does not have a \
                            a qumada_pulse method. Buffered measurements without \
                            ramp method are no longer supported. \
                            Use the unbuffered script!"
                    )
                    raise ex

            if trigger_type == "manual":
                logger.warning(
                    "You are using manual triggering. If you want to pulse parameters on multiple"
                    "instruments this can lead to delays and bad timing!"
                )

            if trigger_type == "hardware":
                try:
                    trigger_start()
                except NameError as ex:
                    print("Please set a trigger or define a trigger_start method")
                    raise ex

            elif trigger_type == "software":
                for buffer in self.buffers:
                    buffer.force_trigger()
                logger.warning(
                    "You are using software trigger, which \
                    can lead to significant delays between \
                    measurement instruments! Only recommended\
                    for debugging."
                )
            timeout_timer = 0
            while not all(buffer.is_finished() for buffer in list(self.buffers)):
                timeout_timer += 0.1
                sleep(0.1)
                if timeout_timer >= buffer_timeout_multiplier * self._burst_duration:
                    raise TimeoutError
            try:
                trigger_reset()
            except TypeError:
                logger.info("No method to reset the trigger defined.")

            results = self.readout_buffers()

            datasaver.add_result(
                (timer, time_setpoints),
                *(zip(self.dynamic_channels, setpoints)),
                *(zip(self.active_compensating_channels, compensating_setpoints)),
                *results,
                *static_gettables,
            )
        datasets.append(datasaver.dataset)
        self.clean_up()
        return datasets


class Generic_Pulsed_Repeated_Measurement(MeasurementScript):
    """
    Executes a buffered pulsed measurement with repeated acquisitions.
    Results of the acquisitions are averaged. All setpoint arrays need to have
    the same length.

    Triggering methods:

    - "software": Sends a software command to each buffer and dynamic parameters
                   to start data acquisition and ramping. Timing might be slightly off.
    - "hardware": Runs trigger_start to start the measurement. Can be preconfigured
                   or manually adjusted (requires `trigger_start` callable).
    - "manual": Trigger setup is user-defined, useful for synchronized trigger outputs.

    Parameters
    ----------
    trigger_start : str or callable, optional
        A callable to start the measurement or "manual" for user-triggered setup.
        Default is "manual".
    trigger_reset : callable, optional
        A callable to reset the trigger after measurement. Default is None.
    trigger_type : str, optional
        Type of trigger to use ("software", "hardware", "manual"). Default is "software".
    include_gate_name : bool, optional
        If True, appends the name of the ramped gates to the measurement name. Default is True.
    repetitions : int, optional
        Number of repeated measurements to perform. Default is 1.
    sync_trigger : callable, optional
        Method for synchronized triggering. Default is None.

    Returns
    -------
    datasets : list of qcodes.dataset.data_set.DataSet
        A list of datasets containing the measurement results.

    Raises
    ------
    AttributeError
        If a required method (e.g., for pulsing) is missing.
    Exception
        If static or dynamic parameters have invalid configurations.

    Notes
    -----
    - Results are averaged across repetitions.s<
    """

    def run(self):
        self.buffered = True
        TRIGGER_TYPES = ["software", "hardware", "manual"]
        trigger_start = self.settings.get("trigger_start", "manual")  # TODO: this should be set elsewhere
        trigger_reset = self.settings.get("trigger_reset", None)
        trigger_type = _validate_mapping(
            self.settings.get("trigger_type"),
            TRIGGER_TYPES,
            default="software",
            default_key_error="software",
        )
        self.repetitions = self.settings.get("repetitions", 1)
        include_gate_name = self.settings.get("include_gate_name", True)
        sync_trigger = self.settings.get("sync_trigger", None)
        datasets = []
        timer = ElapsedTimeParameter("time")
        self.generate_lists()
        self.measurement_name = naming_helper(self, default_name="nD Sweep")
        if include_gate_name:
            gate_names = [gate["gate"] for gate in self.dynamic_parameters]
            self.measurement_name += f" {gate_names}"

        meas = Measurement(name=self.measurement_name)
        meas.register_parameter(timer)
        for parameter in self.dynamic_parameters:
            self.properties[parameter["gate"]][parameter["parameter"]]["_is_triggered"] = True
        for dynamic_param in self.dynamic_channels:
            meas.register_parameter(
                dynamic_param,
                setpoints=[
                    timer,
                ],
            )

        # -------------------
        static_gettables = []
        del_channels = []
        del_params = []
        for parameter, channel in zip(self.gettable_parameters, self.gettable_channels):
            if is_bufferable(channel):
                meas.register_parameter(
                    channel,
                    setpoints=[
                        timer,
                    ],
                )
            elif channel in self.static_channels:
                del_channels.append(channel)
                del_params.append(parameter)
                meas.register_parameter(
                    channel,
                    setpoints=[
                        timer,
                    ],
                )
                parameter_value = self.properties[parameter["gate"]][parameter["parameter"]]["value"]
                static_gettables.append((channel, [parameter_value for _ in range(self.buffered_num_points)]))
        for channel in del_channels:
            self.gettable_channels.remove(channel)
        for param in del_params:
            self.gettable_parameters.remove(param)
        # --------------------------

        self.initialize()
        for c_param in self.active_compensating_channels:
            meas.register_parameter(
                c_param,
                setpoints=[
                    timer,
                ],
            )

        instruments = {param.root_instrument for param in self.dynamic_channels}
        time_setpoints = np.linspace(0, self._burst_duration, int(self.buffered_num_points))
        setpoints = [sweep.get_setpoints() for sweep in self.dynamic_sweeps]
        compensating_setpoints = []
        for i in range(len(self.active_compensating_channels)):
            index = self.compensating_channels.index(self.active_compensating_channels[i])
            active_setpoints = sum([sweep.get_setpoints() for sweep in self.compensating_sweeps[i]])
            active_setpoints += float(self.compensating_parameters_values[index])
            compensating_setpoints.append(active_setpoints)
            if min(active_setpoints) < min(self.compensating_limits[index]) or max(active_setpoints) > max(
                self.compensating_limits[index]
            ):
                raise Exception(f"Setpoints of compensating gate {self.compensating_parameters[index]} exceed limits!")
        results = []
        with meas.run() as datasaver:
            for k in range(self.repetitions):
                self.initialize()
                try:
                    trigger_reset()
                except TypeError:
                    logger.info("No method to reset the trigger defined.")
                self.ready_buffers()
                for instr in instruments:
                    try:
                        instr._qumada_pulse(
                            parameters=[*self.dynamic_channels, *self.active_compensating_channels],
                            setpoints=[*setpoints, *compensating_setpoints],
                            delay=self._burst_duration / self.buffered_num_points,
                            sync_trigger=sync_trigger,
                        )
                    except AttributeError as ex:
                        logger.error(
                            f"Exception: {instr} probably does not have a \
                                a qumada_pulse method. Buffered measurements without \
                                ramp method are no longer supported. \
                                Use the unbuffered script!"
                        )
                        raise ex

                if trigger_type == "manual":
                    logger.warning(
                        "You are using manual triggering. If you want to pulse parameters on multiple"
                        "instruments this can lead to delays and bad timing!"
                    )

                if trigger_type == "hardware":
                    try:
                        trigger_start()
                    except NameError as ex:
                        print("Please set a trigger or define a trigger_start method")
                        raise ex

                elif trigger_type == "software":
                    for buffer in self.buffers:
                        buffer.force_trigger()
                    logger.warning(
                        "You are using software trigger, which \
                        can lead to significant delays between \
                        measurement instruments! Only recommended\
                        for debugging."
                    )

                while not all(buffer.is_finished() for buffer in list(self.buffers)):
                    sleep(0.1)
                try:
                    trigger_reset()
                except TypeError:
                    logger.info("No method to reset the trigger defined.")

                results.append(self.readout_buffers())
            average_results = []
            for i in range(len(results[0])):
                helper_array = np.zeros(len(results[0][0][1]))
                for meas_results in results:
                    helper_array += meas_results[i][1]
                helper_array /= self.repetitions
                average_results.append((meas_results[i][0], helper_array))

            datasaver.add_result(
                (timer, time_setpoints),
                *(zip(self.dynamic_channels, setpoints)),
                *(zip(self.active_compensating_channels, compensating_setpoints)),
                *average_results,
                *static_gettables,
            )
        datasets.append(datasaver.dataset)
        self.clean_up()
        return datasets



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Till Huckeman
# - Daniel Grothe


# Python program to create
# a file explorer in Tkinter

# import filedialog module
import tkinter
from os import curdir
from tkinter import filedialog


# Function for opening the file explorer window
def browsefiles(**kwargs):
    """
    Opens gui for selecting files, returns filepath+name.
    kwargs:
        initialdir ([str], def: "/"): directory to start at
        filetypes ([tuple([str:label],[str: suffix])], def: txt and all files):
            Selectable filetypes
    """
    print("A popup window opened, it is possibly hidden behind other windows...")
    initialdir = kwargs.get("initialdir", curdir)
    filetypes = kwargs.get("filetypes", (("Text files", "*.txt*"), ("all files", "*.*")))
    # Make a top-level instance and hide since it is ugly and big.
    root = tkinter.Tk()
    root.withdraw()
    # Make it almost invisible - no decorations, 0 size, top left corner.
    root.overrideredirect(True)
    root.geometry("0x0+0+0")

    # Show window again and lift it to top so it can get focus,
    # otherwise dialogs will end up behind the terminal.
    root.deiconify()
    root.tkraise()
    root.focus_force()
    filename = filedialog.askopenfilename(
        parent=root, initialdir=initialdir, title="Select a File", filetypes=filetypes
    )
    root.destroy()
    return filename


def browsesavefile(**kwargs):
    """
    Opens gui for creating new file for saving stuff. Returns opened file
    Keep in mind to close it after writing.
    kwargs:
        initialdir ([str], def: "/"): directory to start at
        filetypes ([tuple([str:label],[str: suffix])], def: txt and all files):
            Selectable filetypes

    """
    initialdir = kwargs.get("initialdir", curdir)
    filetypes = kwargs.get("filetypes", (("Text files", "*.txt*"), ("all files", "*.*")))
    # Make a top-level instance and hide since it is ugly and big.
    root = tkinter.Tk()
    root.withdraw()
    # Make it almost invisible - no decorations, 0 size, top left corner.
    root.overrideredirect(True)
    root.geometry("0x0+0+0")

    # Show window again and lift it to top so it can get focus,
    # otherwise dialogs will end up behind the terminal.
    root.deiconify()
    root.tkraise()
    root.focus_force()
    file = tkinter.filedialog.asksaveasfile(parent=root, initialdir=initialdir, filetypes=filetypes)
    root.destroy()
    return file



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe
# - Tobias Hangleiter


from __future__ import annotations

import json
import operator
from collections.abc import Mapping
from os import PathLike
from pathlib import Path

import numpy as np
import pandas as pd


class ParameterDict(dict):
    """Changes operator | and |= of dict to overwrite entries, that are dict and have a key 'type'."""

    def __or__(self, other):
        new = self.__class__(**self)
        new |= other
        return new

    def __ior__(self, other):
        for key, val in other.items():
            # All low-level parameters have an entry "keys".
            # Do not merge but overwrite those entries.
            if isinstance(val, Mapping) and "type" not in val.keys():
                self[key] = operator.ior(self.get(key, ParameterDict()), val)
            else:
                self[key] = val
        return self


def update_parameters(parameter_file_path: str | PathLike, new_parameters: ParameterDict):
    """Write or overwrite parameters to a file."""
    path = Path(parameter_file_path)
    existing_parameters = json.load(path.open("r" if path.exists() else "x"), object_pairs_hook=ParameterDict)
    parameters = existing_parameters | new_parameters
    json.dump(parameters, path.open("w"), indent=2)
    return parameters


def intialize_dac_parameter_file(mapping, parameter_file_path: str | PathLike):
    """Write new dac parameter file with dummy channels before relevant entries."""
    parameters = ParameterDict()
    for i in range(20):
        if i in mapping.keys():
            key = mapping[i]
        else:
            key = f"CH{i}"
        parameters |= gettable_dac_entry(key)

    return update_parameters(parameter_file_path, parameters)


def excel_to_dac_parameter_file(
    excel_file,
    parameter_file_path: str | PathLike,
    sample_header="Sample",
    dac_header="DAC/AWG",
):
    """Read dac parameters from Excel file and save them (with dummy channels) to new parameter file."""
    excel = pd.read_excel(excel_file, usecols=[sample_header, dac_header])
    mapping = ParameterDict(
        {int(dac): sample for dac, sample in zip(excel["DAC/AWG"], excel["Sample"]) if not np.isnan(dac)}
    )

    return intialize_dac_parameter_file(mapping, parameter_file_path)


def dynamic_dac_entry(dac, start, stop, num_points=100, delay=0.025) -> ParameterDict:
    """Generate a "dynamic" dac parameter."""
    return ParameterDict(
        {
            dac: ParameterDict(
                {
                    "voltage": ParameterDict(
                        {"type": "dynamic", "start": start, "stop": stop, "num_points": num_points, "delay": delay}
                    )
                }
            )
        }
    )


def static_dac_entry(dac, value, gettable=False) -> ParameterDict:
    """Generate a "static" dac parameter."""
    return ParameterDict(
        {
            dac: ParameterDict(
                {"voltage": ParameterDict({"type": "static gettable" if gettable else "static", "value": value})}
            )
        }
    )


def gettable_dac_entry(dac, **kwargs) -> ParameterDict:
    """Generate a "gettable" dac parameter."""
    return ParameterDict({dac: ParameterDict({"voltage": ParameterDict({"type": "gettable"}) | kwargs})})


def static_smu_entry(smu, parameter, value, gettable=False) -> ParameterDict:
    """Generate a static (gettable) SMU parameter."""
    return ParameterDict(
        {
            smu: ParameterDict(
                {parameter: ParameterDict({"type": "static gettable" if gettable else "static", "value": value})}
            )
        }
    )



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Bertha Lab Setup
# - Daniel Grothe
# - Till Huckeman


import threading

# most of the drivers only need a couple of these... moved all up here for clarity below
from time import sleep

import numpy as np
from qcodes.instrument import Instrument
from qcodes.parameters import Parameter
from qcodes.validators import validators as vals


# %%
class dmm_results_random(Parameter):
    def get_raw(self):
        return np.random.sample()


class dmm_results_sinus(Parameter):
    def get_raw(self):
        length = self.root_instrument.buffer_n_points()
        if length and length > 0:
            return [
                np.sin(2 * np.pi * x / (self.root_instrument.buffer_n_points() ** 2 / self.root_instrument.buffer_SR()))
                for x in np.linspace(0, length, length)
            ]
        else:
            raise Exception("Set buffer_n_points first to a positive value")


class dmm_buffer(Parameter):
    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)
        self.buffer_data = []
        self.buffer_length = 512
        self.SR = 512
        self.is_finished = True
        self.subscribed_params = list()
        self.triggered: bool = False
        self._is_triggered = self.root_instrument._trigger_event

    def subscribe(self, param):
        assert param.root_instrument == self.root_instrument
        if param not in self.subscribed_params:
            self.subscribed_params.append(param)

    def ready_buffer(self):
        self.SR = self.root_instrument.buffer_SR()
        self.buffer_length = self.root_instrument.buffer_n_points()
        self.buffer_data = [list() for _ in self.subscribed_params]
        self.is_finished = False

    def start(self):
        if self.is_finished:
            raise Exception("Buffer is not ready!")
        self.thread = threading.Thread(target=self._run, args=(), daemon=True)
        self.thread.start()

    def _run(self):
        _ = self._is_triggered.wait()
        for i in range(0, self.buffer_length):
            for j in range(len(self.subscribed_params)):
                datapoint = self.subscribed_params[j]()
                if isinstance(datapoint, float):
                    self.buffer_data[j].append(self.subscribed_params[j]())
                elif isinstance(datapoint, list):
                    self.buffer_data[j].append(self.subscribed_params[j]()[i])
            sleep(1 / self.SR)
        self.is_finished = True

    def reset(self):
        self.buffer_data = []

    def get_raw(self):
        return self.buffer_data


class DummyDmm(Instrument):
    def __init__(self, name, trigger_event=threading.Event(), **kwargs):
        super().__init__(name, **kwargs)

        self._trigger_event = trigger_event

        self.add_parameter(
            "voltage",
            unit="V",
            parameter_class=dmm_results_sinus,
        )

        self.add_parameter("current", unit="A", parameter_class=dmm_results_random)

        self.add_parameter("buffer", unit="V", parameter_class=dmm_buffer)

        self.add_parameter(
            "buffer_SR",
            unit="Sa/s",
            set_cmd=None,
            vals=vals.Numbers(0, 512),
        )

        self.add_parameter(
            "trigger_mode",
        )

        self.add_parameter("is_finished", get_cmd=self._is_finished)

        self.add_parameter("buffer_n_points", set_cmd=None, vals=vals.Ints(0, 16383))

        self.add_parameter("triggered", set_cmd=None, vals=vals.Bool())
        self.triggered(False)

        self.add_function("start", call_cmd=self._start_buffer)

        self.add_function("ready_buffer", call_cmd=self._ready_buffer)

        self.add_function("reset_buffer", call_cmd=self._reset_buffer)

    def _force_trigger(self):
        self.buffer._is_triggered.set()
        return None

    def _start_buffer(self):
        print("Started buffer")
        self.buffer.start()
        return None

    def _ready_buffer(self):
        print("Buffer is now ready")
        self.buffer.ready_buffer()
        self.buffer._is_triggered = self._trigger_event
        return None

    def _reset_buffer(self):
        print("Buffer was resetted")
        self.buffer.buffer_data = []
        return None

    def _is_finished(self):
        return self.buffer.is_finished
